from odoo import models, fields, api
from odoo.tools import html2plaintext

from random import randint

import logging
_logger = logging.getLogger(__name__)


class resUsers(models.Model):
    _inherit = 'res.users'

    def action_update_right(self):
        self.ensure_one()
        wz = self.env['access.control.wizard'].create({
            'user_id': self.id,
        })
        wz.next()
        return wz.reopen()


class accessControlTemplate(models.Model):
    _name = 'access.control.template'
    _description = 'Template of access rights'
    _rec_names_search = ['name', 'code']

    code = fields.Char(required=True)
    name = fields.Char(required=True)
    group_ids = fields.Many2many('res.groups')

    def copy(self, default=None):
        self.ensure_one()
        default = dict(default or {}, code='%s_%.3d' % (self.code, randint(100, 999)))
        return super().copy(default)

    @api.depends('code', 'name')
    def _compute_display_name(self):
        for rec in self:
            rec.display_name = "[%s] %s" % (rec.code, rec.name)

    @api.model
    def apply(self, uids, actions, reason):
        """Method used in RPC by sysadmin.

        uids = list of id of res.users or login
        actions = list of tuple (action, value)
          action in: 'group_add', 'group_remove', 'template_add', 'template_remove', 'template_reset'
          value = template code (if action in 'template_add', 'template_remove', 'template_reset')
          value = group_id or xml_id (if action in 'group_add', 'group_remove')
        """

        if not isinstance(uids, list):
            uids = [uids]
        if not isinstance(actions, list):
            actions = [actions]

        supported_actions = self.env['access.control.wizard']._fields['action_todo'].get_values(self.env)
        for (action, _val) in actions:
            assert action in supported_actions, "Invalid action %s" % action

        resp = []
        for uid in uids:
            if not str(uid).isdigit():
                login = str(uid)
                if not login.endswith('@odoo.com'):
                    login = "%s%s" % (uid, '@odoo.com')
                user = self.env['res.users'].search([('login', '=', login)])
                assert len(user) == 1, "Login <%s> match %s res.users" % (login, len(user))
            else:
                user = self.env['res.users'].browse(uid)
                assert '@odoo.com' in user.login, "Uid <%s> doesn't match Odoo User" % uid

            wzr = self.env['access.control.wizard'].create({
                'user_id': user.id,
                'reason': reason,
            })

            for (action, value) in actions:
                wzr.action_todo = action
                if action.startswith('template_'):
                    wzr.template_id = self.search([('code', '=', value)]).id
                    assert wzr.template_id, "Invalid value for template %s" % value
                else:
                    if str(value).isdigit():
                        wzr.group_id = value
                    else:
                        g = self.env.ref(value)
                        assert g._name == 'res.groups'
                        wzr.group_id = g.id
                    assert wzr.group_id, "Invalid value for group %s" % value
                wzr.simulate()

            wzr.apply_rights()
            resp.append(html2plaintext(str(wzr.log)))
        return resp

    _sql_constraints = [
        ('code_uniq', 'unique (code)', 'The code should be unique !')
    ]


class accessControlUpdate(models.Model):
    _name = 'access.control.log'
    _description = 'Access Control Log'
    _order = 'id desc'

    ip = fields.Char(string='IP')
    user_id = fields.Many2one('res.users', string='Modificator', help="Author of modification")
    target_user_id = fields.Many2one('res.users', string='Target User', help="User modified", index=True)
    reason = fields.Text(string='Action Reason', help="Reason of modification")
    action = fields.Selection(string="Action", selection=[('group_update', 'Update Group'), ('company_update', 'Update Companies')])
    details = fields.Text(string='Details', help="Resume of changes")

    def _compute_display_name(self):
        actions_labels = dict(self._fields['action'].selection)
        for rec in self:
            rec.display_name = "%s: %s" % (actions_labels[rec.action], rec.reason[:250])


class accessControlWizardGroup(models.TransientModel):
    _name = 'access.control.wizard.group'
    _description = 'Line of Access Control Wizard'

    wizard_id = fields.Many2one('access.control.wizard')
    group_id = fields.Many2one('res.groups', required=True, string="Group")
    name = fields.Char(related='group_id.full_name')
    group_name = fields.Char(related='group_id.name', string="Real Group")
    category_name = fields.Char(related='group_id.category_id.name', string="Category")
    add = fields.Boolean()
    remove = fields.Boolean()
    ingroup = fields.Boolean()

    def add_group(self):
        if not self.ingroup:
            self.add = True
        else:
            self.remove = False
        return self.wizard_id.reopen()

    def remove_group(self):
        wzr = self.wizard_id
        if self.ingroup:
            self.remove = True
        else:
            self.unlink()
        return wzr.reopen()


class accessControlWizard(models.TransientModel):
    _name = 'access.control.wizard'
    _description = 'Access Control Wizard'

    user_id = fields.Many2one('res.users', required=True, domain="[('share', '=', False)]")
    action_todo = fields.Selection([
        ('group_add', 'Add a group'),
        ('group_remove', 'Remove a group'),
        ('template_add', 'Add a template'),
        ('template_remove', 'Remove a template'),
        ('template_reset', 'Reset to template'),
    ], default='template_reset', required=True)
    template_id = fields.Many2one('access.control.template')
    group_id = fields.Many2one('res.groups', string="Group")

    group_ids = fields.One2many(comodel_name='access.control.wizard.group', inverse_name='wizard_id', string="Groups")
    group_ids_display = fields.One2many(comodel_name='access.control.wizard.group', compute="filter_group_ids")
    group_ids_display_simplify = fields.One2many(comodel_name='access.control.wizard.group', compute="filter_group_ids")

    reason = fields.Text(required=False)
    warn = fields.Text(compute='get_warn')
    log = fields.Html(readonly=True)

    # filter / ux
    hide_technical = fields.Boolean(default=True)
    hide_ingroup = fields.Boolean(default=False)
    mode_advanced = fields.Boolean(default=1)
    step = fields.Integer(default=0)

    @api.depends('group_ids', 'hide_technical', 'hide_ingroup')
    def filter_group_ids(self):
        def is_visible(rec, hide_ingroup, hide_technical):
            hidden_group = [self.env.ref(gr) for gr in self.env['res.groups']._get_hidden_extra_categories()]
            hide = False
            if not (rec.add or rec.remove):
                if hide_technical and rec.group_id.category_id in hidden_group:
                    hide = True
                if hide_ingroup and not (rec.add or rec.remove):
                    hide = True
            return hide

        groups_by_app = self.env['res.groups'].get_groups_by_application()
        groups_by_app.sort(key=lambda gba: gba[3][0])
        for rec in self:
            rec.group_ids_display = self.group_ids.filtered(
                lambda x: not is_visible(x, self.hide_ingroup, self.hide_technical)
            )
            rec.group_ids_display_simplify = False
            for (_cat, typ, groups, (_seq, _cat_name)) in groups_by_app:
                if typ == 'selection':
                    for g in groups[::-1]:
                        find = [wg for wg in rec.group_ids if g in wg.group_id]
                        if find:
                            rec.group_ids_display_simplify += find[0]
                            break
        return self.reopen()

    def simulate(self):
        if self.action_todo == 'group_add' and self.group_id:
            self._add_groups([self.group_id])
        elif self.action_todo == 'group_remove' and self.group_id:
            self._remove_groups(self.group_id)
        elif self.action_todo == 'template_add':
            self._add_groups(self.template_id.group_ids)
        elif self.action_todo == 'template_remove':
            self._remove_groups(self.template_id.group_ids)
        elif self.action_todo == 'template_reset':
            current_groups = self.group_ids.group_id
            futur_groups = self.template_id.group_ids + self._get_default_groups()
            self._add_groups(futur_groups - current_groups)
            self._remove_groups(current_groups - futur_groups)
        return self.reopen()

    def _get_default_groups(self):
        # group_hr_holidays_responsible is on default user but removed by _clean_leave_responsible_users
        hr_responsible = self.env.ref('hr_holidays.group_hr_holidays_responsible')
        default_user = self.env.ref('openerp_enterprise.access_control_default_user', raise_if_not_found=False)
        return (default_user or self.env.ref('base.default_user', raise_if_not_found=False)).groups_id - hr_responsible

    def _add_groups(self, groups_to_add):
        current_groups = self.group_ids.mapped(lambda g: (g.group_id, g.id))
        for g in groups_to_add:
            if g not in dict(current_groups):
                self.env['access.control.wizard.group'].create({
                    'wizard_id': self.id,
                    'group_id': g.id,
                    'add': True,
                    'remove': False,
                    'ingroup': False,
                })

    def _remove_groups(self, groups_to_remove):
        current_groups = dict(self.group_ids.mapped(lambda g: (g.group_id, g)))
        for g in (groups_to_remove - self._get_default_groups()):
            if g in current_groups:
                if current_groups[g].ingroup:
                    current_groups[g].remove = True
                    current_groups[g].add = False
                else:
                    current_groups[g].unlink()

    def apply_rights(self):
        adds = self.group_ids.filtered('add')
        removes = self.group_ids.filtered('remove')
        new_groups = self.user_id.groups_id + adds.mapped('group_id') - removes.mapped('group_id')
        self.with_context(update_reason=self.reason).user_id.groups_id = new_groups
        self.step += 1
        self.log = """
            <p><b>UPDATE rights for: </b>{login}</p>
            <p>
                <b>Add: </b>
                <ul>{add}</ul>
            </p>
            <p>
                <b>Remove: </b>
                <ul>{remove}</ul>
            </p>
            <p>
                <b>Reason:</b><br/>
                <i>{reason}</i>
            </p>
        """.format(
            login=self.user_id.login.upper(),
            add=''.join(["<li>%s</li>" % g for g in adds.mapped('name')]),
            remove=''.join(["<li>%s</li>" % g for g in removes.mapped('name')]),
            reason=self.reason,
        )
        return self.reopen()

    # UX
    def reopen(self):
        return {
            'type': 'ir.actions.act_window',
            'name': 'Manage Rights for: %s' % self.user_id.name,
            'res_model': 'access.control.wizard',
            'view_mode': 'form',
            'view_type': 'form',
            'target': 'new',
            'res_id': self.id,
        }

    def prev(self):
        self.ensure_one()
        self.step -= 1
        return self.reopen()

    def next(self):
        self.ensure_one()
        self.step += 1

        if self.step == 1:
            # reset group based on selected user
            for g in self.user_id.groups_id:
                self.env['access.control.wizard.group'].create({
                    'wizard_id': self.id,
                    'group_id': g.id,
                    'add': False,
                    'remove': False,
                    'ingroup': True,
                })
        return self.reopen()

    def reload_user(self):
        self.ensure_one()
        for g in self.group_ids:
            if g.ingroup:
                g.remove = g.add = False
            else:
                g.unlink()
        return self.reopen()

    @api.depends('group_ids')
    def get_warn(self):
        User = self.env['res.users']
        self.warn = self.user_id and User._prepare_warning_for_group_inheritance(
            User.new({
                'name': self.user_id.name,
                'login': 'WzUser',
                'groups_id': self.group_ids.filtered(lambda x: not x.remove).group_id,
            })
        ) or '-'
